#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define MAXCOLUMNS 128
static struct Column {
	char name[128];
	char type[128];
} column[MAXCOLUMNS];
static int columns=0;

static bool i_used=false;
static bool f_used=false;
static bool r_used=false;
static bool s_used=false;
static bool b_used=false;


static void generate_header(const char *basename, const char *filename) {
	FILE *fp=fopen(filename,"w");
	if(!fp) {
		perror(filename);
		exit(EXIT_FAILURE);
	}
	
	fprintf(fp,"#ifndef _%s_H_\n",basename);
	fprintf(fp,"#define _%s_H_\n",basename);
	fprintf(fp,"\n");
	fprintf(fp,"#include <vector>\n");
	if(i_used || r_used)
		fprintf(fp,"#include <inttypes.h>\n");
	if(r_used || s_used)
		fprintf(fp,"#include <string>\n");
	fprintf(fp,"\n");
	fprintf(fp,"struct %s {\n",basename);
	for(int i=0; i<columns; i++) {
		const char *c_type=0;
		if(strcmp(column[i].type,"i")==0)
			c_type="int64_t";
		if(strcmp(column[i].type,"f")==0)
			c_type="double";
		if(strcmp(column[i].type,"s")==0)
			c_type="std::string";
		if(strcmp(column[i].type,"r")==0)
			c_type="std::string<uint8_t>";
		if(strcmp(column[i].type,"b")==0)
			c_type="bool";
		fprintf(fp,"\t%-16s\t%s;\n",c_type,column[i].name);
	}
	fprintf(fp,"};\n");
	fprintf(fp,"\n");
	fprintf(fp,"typedef std::vector<%s> RowSet_%s;\n",basename,basename);
	fprintf(fp,"\n");
	fprintf(fp,"bool LoadFromFile(const char *filename, RowSet_%s *result);\n",basename);
	fprintf(fp,"bool WriteToFile(const char *filename, const RowSet_%s &rowset);\n",basename);
	fprintf(fp,"\n");
	fprintf(fp,"#endif\n");
	
	fclose(fp);
}


static const char parse_i[]=
"static char *parse_i(char *n, int64_t *value) {\n"
"	char *e=n;\n"
"	if(*e=='-') e++;\n"
"	while(isdigit(*e)) e++;\n"
"	if(*e) *e++ = '\\0';\n"
"	int rc = sscanf(n,\"%\"PRId64, value);\n"
"	if(rc!=1) return 0;\n"
"	return e;\n"
"}\n";

static const char parse_f[]=
"static char *parse_f(char *n, double *value) {\n"
"	char *endptr=0;\n"
"	*value = strtod(n,&endptr);\n"
"	if(endptr) {\n"
"		if(*endptr=='\\0' || isspace(*endptr))\n"
"			return endptr;\n"
"		else\n"
"			return 0;\n"
"	} else {\n"
"		while(*n) n++;\n"
"		return n;\n"
"	}\n"
"}\n";

static const char parse_s[]=
"static char *parse_s(char *n, std::string *value) {\n"
"	if(*n=='\"') {\n"
"		char *p=strchr(n,'\"');\n"
"		if(p) value->reserve(p-n);\n"
"		n++;\n"
"		while(*n) {\n"
"			if(*n=='\\\\') n++;\n"
"			else if(*n=='\"') break;\n"
"			*value += *n++;\n"
"		}\n"
"		if(*n=='\"') return n+1;\n"
"		return n;\n"
"	} else {\n"
"		char *p=n;\n"
"		while(*p && !isspace(*p))\n"
"			p++;\n"
"		value->assign(n,(p-n));\n"
"		return p;\n"
"	}\n"
"}\n";

static const char parse_r[]=
"static char *parse_r(char *n, std::string<uint8_t> *value) {\n"
"	if(n[0]=='0' && n[1]=='x') n+=2;\n"
"	char *e=n;\n"
"	while(isxdigit(*e)) e++;\n"
"	if((e-n)%2) return 0;\n"
"	size_t rl = (e-n)/2;\n"
"	value->reserve(rl);\n"
"	while(rl) {\n"
"		char c1 = *n++;\n"
"		char c2 = *n++;\n"
"		rl--;\n"
"		uint8_t xv1;\n"
"		if(c1>='0' && c1<='9') xv1=c1-'0';\n"
"		else if(c1>='a' && c1<='f') xv1=c1-'a'+10;\n"
"		else xv1=c1-'A'+10;\n"
"		uint8_t xv2;\n"
"		if(c2>='0' && c2<='9') xv2=c2-'0';\n"
"		else if(c2>='a' && c2<='f') xv2=c2-'a'+10;\n"
"		else xv2=c2-'A'+10;\n"
"		uint8_t v = (xv1<<4)|xv2;\n"
"		*value += v;\n"
"	}\n"
"	return e;\n"
"}\n";

static const char parse_b[]=
"static char *parse_b(char *n, bool *value) {\n"
"	char *e=n;\n"
"	while(*e && !isspace(*e)) e++;\n"
"	if(*e) *e++ = '\\0';\n"
"	if(strcmp(n,\"true\")==0 || strcmp(n,\"T\")==0) *value=true;\n"
"	else if(strcmp(n,\"false\")==0 || strcmp(n,\"F\")==0) *value=false;\n"
"	else return 0;\n"
"	return e;\n"
"}\n";


static void generate_implementation(const char *basename, const char *filename) {
	FILE *fp=fopen(filename,"w");
	if(!fp) {
		perror(filename);
		exit(EXIT_FAILURE);
	}
	
	fprintf(fp,"#define __STDC_FORMAT_MACROS\n");
	fprintf(fp,"#include \"%s.h\"\n",basename);
	fprintf(fp,"#include <stdio.h>\n");
	fprintf(fp,"#include <ctype.h>\n");
	if(f_used) fprintf(fp,"#include <stdlib.h>\n");
	if(s_used || b_used) fprintf(fp,"#include <string.h>\n");
	fprintf(fp,"\n");
	if(i_used) fputs(parse_i,fp);
	if(f_used) fputs(parse_f,fp);
	if(r_used) fputs(parse_r,fp);
	if(s_used) fputs(parse_s,fp);
	if(b_used) fputs(parse_b,fp);
	
	fprintf(fp,"bool LoadFromFile(const char *filename, RowSet_%s *result)\n",basename);
	fprintf(fp,"{\n");
	fprintf(fp,"\tFILE *fp=fopen(filename,\"r\");\n");
	fprintf(fp,"\tif(!fp) return false;\n");
	fprintf(fp,"\t\n");
	fprintf(fp,"\tresult->clear();\n");
	
	fprintf(fp,"\t\n");
	
	fprintf(fp,"\tchar line[1024];\n");
	fprintf(fp,"\twhile(fgets(line,sizeof(line),fp)) {\n");
	fprintf(fp,"\t\tchar *n=line;\n");
	fprintf(fp,"\t\t%s new_row;\n",basename);
	fprintf(fp,"\n");
	for(int i=0; i<columns; i++) {
		const Column &c=column[i];
		if(i!=0)
			fprintf(fp,"\t\twhile(isspace(*n)) n++;\n");
		fprintf(fp,"\t\tn = parse_%s(n,&new_row.%s);\n",c.type,c.name);
		fprintf(fp,"\t\tif(!n) {\n");
		fprintf(fp,"\t\t\tfclose(fp);\n");
		fprintf(fp,"\t\t\treturn false;\n");
		fprintf(fp,"\t\t}\n");
		fprintf(fp,"\n");
	}
	fprintf(fp,"\t\tresult->push_back(new_row);\n");
	fprintf(fp,"\t}\n");
	fprintf(fp,"\tfclose(fp);\n");
	fprintf(fp,"\treturn true;\n");
	fprintf(fp,"}\n");
	
	fprintf(fp,"\n");
	
	fprintf(fp,"bool WriteToFile(const char *filename, const RowSet_%s &rowset) {\n",basename);
	fprintf(fp,"\tFILE *fp=fopen(filename,\"w\");\n");
	fprintf(fp,"\tif(!fp) return false;\n");
	fprintf(fp,"\t\n");
	fprintf(fp,"\tfor(RowSet_%s::const_iterator i=rowset.begin();\n",basename);
	fprintf(fp,"\t    i!=rowset.end();\n");
	fprintf(fp,"\t    i++)\n");
	fprintf(fp,"\t{\n");
	for(int i=0; i<columns; i++) {
		const Column &c = column[i];
		if(i!=0) {
			fprintf(fp,"\t\t\n");
		}
		if(strcmp(c.type,"i")==0)
			fprintf(fp,"\t\tfprintf(fp,\"%%\"PRId64,i->%s);\n",c.name);
		else if(strcmp(c.type,"f")==0)
			fprintf(fp,"\t\tfprintf(fp,\"%%f\",i->%s);\n",c.name);
		else if(strcmp(c.type,"s")==0) {
			fprintf(fp,"\t\tif(i->%s.size()!=0 &&\n",c.name);
			fprintf(fp,"\t\t   (i->%s[0]=='\"' || i->%s.find_first_of(' ')!=std::string::npos)) {\n",c.name,c.name);
			fprintf(fp,"\t\t\tfputc('\"',fp);\n");
			fprintf(fp,"\t\t\tfor(size_t j=0; j<i->%s.size(); j++) {\n",c.name);
			fprintf(fp,"\t\t\t\tchar c=i->%s[j];\n",c.name);
			fprintf(fp,"\t\t\t\tif(c=='\"' || c=='\\\\') putc('\\\\',fp);\n");
			fprintf(fp,"\t\t\t\tputc(c,fp);\n");
			fprintf(fp,"\t\t\t}\n");
			fprintf(fp,"\t\t\tfputc('\"',fp);\n");
			fprintf(fp,"\t\t} else\n");
			fprintf(fp,"\t\t\tfprintf(fp,\"%%*.*s\",(int)i->%s.size(),(int)i->%s.size(),i->%s.data());\n",c.name,c.name,c.name);
		} else if(strcmp(c.type,"r")==0) {
			fprintf(fp,"\t\tif(i->%s.size()==0) fprintf(fp,\"0x\");\n",c.name);
			fprintf(fp,"\t\tfor(size_t j=0; j<i->%s.size(); j++)\n",c.name);
			fprintf(fp,"\t\t\tfprintf(fp,\"%%02\"PRIx8,i->%s[j]);\n",c.name);
		} else if(strcmp(c.type,"b")==0) {
			fprintf(fp,"\t\tfprintf(fp,i->%s?\"true\":\"false\");\n",c.name);
		}
		
		if(i<columns-1)
			fprintf(fp,"\t\tfputc(' ',fp);\n");
		else {
			fprintf(fp,"\t\t\n");
			fprintf(fp,"\t\tif(fputc('\\n',fp)==EOF) {;\n");
			fprintf(fp,"\t\t\tfclose(fp);\n");
			fprintf(fp,"\t\t\treturn false;\n");
			fprintf(fp,"\t\t}\n");
		}
	}
	fprintf(fp,"\t}\n");
	fprintf(fp,"\tfclose(fp);\n");
	fprintf(fp,"\treturn true;\n");
	fprintf(fp,"}\n");
	fclose(fp);
}


static void printUsage(FILE *fp) {
	fprintf(fp,"usage: gen_ffdb tablename fielddef1 [fielddef2 ...]\n");
	fprintf(fp,"  tablename  base name of the table and used for generated filenames\n"
	           "             and structure names\n"
	           "  fielddef   A field definition in the form name:type\n"
	           "             where <type> can be :\n"
	           "               i (integer)\n"
	           "               f (floating point)\n"
	           "               s (string)\n"
	           "               r (raw)\n"
	           "               b (boolean\n"
	        );
}


int main(int argc, char **argv) {
	if(argc<3) {
		printUsage(stdout);
		return EXIT_FAILURE;
	}
	
	for(int i=2; i<argc; i++) {
		char *ss;
		char *name = strtok_r(argv[i],":",&ss);
		if(!name) {
			fprintf(stderr,"Field %d not in proper form\n",i-1);
			printUsage(stderr);
			return EXIT_FAILURE;
		}
		char *type = strtok_r(0,"",&ss);
		if(!type) {
			fprintf(stderr,"Field %d not in proper form\n",i-1);
			printUsage(stderr);
			return EXIT_FAILURE;
		}
		if(strcmp(type,"i")==0) i_used=true;
		else if(strcmp(type,"f")==0) f_used=true;
		else if(strcmp(type,"s")==0) s_used=true;
		else if(strcmp(type,"r")==0) r_used=true;
		else if(strcmp(type,"b")==0) b_used=true;
		else {
			fprintf(stderr,"Unknown field type %s\n",type);
			return EXIT_FAILURE;
		}
		
		strcpy(column[columns].name,name);
		strcpy(column[columns].type,type);
		columns++;
	}
	
	char filename[256];
	sprintf(filename,"%s.h",argv[1]);
	generate_header(argv[1],filename);
	sprintf(filename,"%s.cc",argv[1]);
	generate_implementation(argv[1],filename);
	
	return 0;
}
